(*=======================================================================================*)				
(*																						 *)
(*							gestion de files de priorit								 *)
(*						utilisation d'arbres tournois parfaits							 *)
(*									 ou heaps											 *)
(*																						 *)
(*=======================================================================================*)				

type 'a structure_de_tas = Tas of ('a vect * (unit -> int) * (int -> unit) * ('a -> 'a -> bool) * int) ;;

(* un tas est un tableau, une fonction qui renvoie sa sa_taille, *)
(* une fonction qui incr_taillemente sa sa_taille, la fonction <,	  *)
(* et la sa_taille maximale admissible pour le tas				  *)

exception Tas_Plein ;;
exception Tas_Vide ;;

let change_dans_vect t i j =
	let temp = t.(i) in t.(i) <- t.(j) ; t.(j) <- temp ;;
	
(* descend_dans_tas t i fait descendre  sa place dans le tas t *)
(* l'lment d'indice i *)

let rec descend_dans_tas t i =
	match t with Tas(tas,sa_taille,_,inf,_) ->
	let n = sa_taille ()
	in
	let fils = 2*i + 1
	in
	let fils = if (fils < n-1) & (inf tas.(fils+1) tas.(fils)) then fils+1 else fils
	in
	if fils < n & inf tas.(fils) tas.(i) then
		begin
			change_dans_vect tas i fils ;
			descend_dans_tas t fils
		end ;;

let rec remonte_dans_tas t i =
	match t with Tas(tas,_,_,inf,_) ->
	if i > 0 & inf tas.(i) tas.((i-1)/2) then
	begin
		change_dans_vect tas i ((i-1)/2) ;
		remonte_dans_tas t ((i-1)/2)
	end ;;

let insre_dans_tas t x =
	match t with Tas(tas,sa_taille,incr_taille,inf,nMAX) ->
	let n = sa_taille ()
	in
	if n = nMAX then raise Tas_Plein ;
	tas.(n) <- x ;
	incr_taille 1 ;
	remonte_dans_tas t n ;;
	
let extrait_minimum t =
	match t with Tas(tas,sa_taille,incr_taille,_,_) ->
	let n = sa_taille ()
	in
	if n = 0 then raise Tas_Vide ;
	let mini = tas.(0)
	in
	begin
		change_dans_vect tas 0 (n-1) ;
		incr_taille (-1) ;
		descend_dans_tas t 0 ;
		mini
	end ;;
	
let tas_en_liste t =
    match t with Tas(tas,sa_taille,_,_,_) -> 
    let n = sa_taille ()
    in
    let rec puise nb l = if nb = 0 then l else puise (nb-1) (tas.(nb-1) :: l)
    in puise n [] ;;
    
let tas_en_vecteur t = vect_of_list (tas_en_liste t) ;;
    
let percolation = remonte_dans_tas ;;
	
(* on fournit la relation d'ordre inf, une liste NON VIDE d'objets, *)
(* et un majorant de la sa_taille du tas *)
(* on rend une structure de tas *)

let liste_en_tas inf (a::q as l) nMAX =
	let n = list_length l
	in
	let sa_taille = ref n
	in
	let tas = make_vect nMAX a
	in
	let rec copie_liste i = function
		  [] -> ()
		| a::q -> tas.(i) <- a ; copie_liste (i+1) q
	in
	copie_liste 0 l ;
	let t = Tas (	tas,
					(function () -> !sa_taille),
					(function delta -> sa_taille := !sa_taille + delta),
					inf,
					nMAX 	)
	in
	for i = (n/2 -1) downto 0 do descend_dans_tas t i done ;
	t ;;
